# -*- coding: utf-8 -*-
# from .inversion import plot, newWindow

#print(f"Invoking __init__.py for {__name__}")

from .inver import inversion
from .potential_prism import Prism, Prism_calc

__all__ = ["inversion", "Prism", "Prism_calc"]